import {
  ErrorCode,
  Notification,
  Request,
  Response,
} from './protocol';

export type OnDisconnectFunc = () => void;
export type OnResponseFunc = (response: any) => void;

const PROTOCOL_NAME = '1.0';

type ID_TYPE = number;

interface RequestQueue {
  resolve: (value?: any) => void;
  reject: (reason?: any) => void;
}

enum WebSocketStatus {
  NOT_CONNECT = 1,
  WILL_CONNECT,
  CONNECTED,
  WILL_DISCONNECT,
}

enum WebSocketReadyState {
  CONNECTING = 0,
  OPEN = 1,
  CLOSING = 2,
  CLOSED = 3,
}

export default class WebSocketRPC {
  private url: string;
  private connStatus: WebSocketStatus;
  private conn: WebSocket;
  private isReady: boolean;
  private requestID: ID_TYPE = 0;

  private onDisconnectListeners: OnDisconnectFunc[] = [];
  private onResponseListeners: Map<string, OnResponseFunc[]>;

  private requestQueue: Map<ID_TYPE, RequestQueue>;

  /**
   * name
   */
  public constructor(url: string) {
    this.isReady = false;
    this.connStatus = WebSocketStatus.NOT_CONNECT;
    this.url = url;
    this.onResponseListeners = new Map();
    this.requestQueue = new Map();
  }

  public Call(method: string, args: any): Promise<any> {
    return new Promise<any>((resolve, reject) => {
      const requestID = this.getRequestID();
      let request = new Request<ID_TYPE>(requestID, method, args);
      request.Protocol = PROTOCOL_NAME;

      this.conn.send(JSON.stringify(request));
      this.requestQueue.set(requestID, {resolve: resolve, reject: reject});
    });
  }

  public OnDisconnect(cb: OnDisconnectFunc): void {
    this.onDisconnectListeners.push(cb);
  }

  public initialize(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this.conn = new WebSocket(this.url);
      } catch(e) {
        console.log(e);
      }
      this.connStatus = WebSocketStatus.WILL_CONNECT;
      this.conn.onopen = (ev: Event): any => {
        this.connStatus = WebSocketStatus.CONNECTED;
        this.attachCallback();
        resolve();
        this.isReady = true;
        return null;
      };

      this.conn.onerror = (ev: Event): any => {
        this.connStatus = WebSocketStatus.NOT_CONNECT;
        reject(new Error('Cannot connect to API Server.'));
        return null;
      };
    });
  }

  private attachCallback(): void {
    this.conn.onopen = (ev: Event): any => {
      this.connStatus = WebSocketStatus.CONNECTED;
      this.isReady = true;
      return null;
    };

    this.conn.onclose = (ev: CloseEvent): any => {
      this.connStatus = WebSocketStatus.NOT_CONNECT;
      this.fireOnDisconnect();
      return null;
    };

    this.conn.onmessage = (ev: MessageEvent): any => {
      const json = JSON.parse(ev.data);
      try {
        let response: Response<ID_TYPE> = Response.convert(json);
        this.onResponseHandler(response);
      } catch (e) {
        // Check message is Notification
        try {
          let notification = Notification.convert(json);
          this.onNotificationHandler(notification);
        } catch (e) {
          console.log(e);
        }
      }

      this.fireOnMessage(ev);
    };

    this.conn.onerror = (ev: Event): any => {
      console.log(ev);
      return null;
    };
  }

  private onResponseHandler(response: Response<ID_TYPE>): void {
      const requestID = response.ID;
      let error = response.Error;
      let result;

      if (undefined !== response.Result) {
        result = JSON.parse(response.Result);
      }

      if (this.requestQueue.has(requestID)) {
        let promise = this.requestQueue.get(requestID);
        this.requestQueue.delete(requestID);
        if (null != result) {
          console.log(result);
          promise.resolve(result);
        } else if (null != error) {
          if (ErrorCode.E_INTERNAL === error.Code) {
            if (undefined !== error.Message) {
              const msgs = error.Message.split('|');
              let errorName: string;
              let errorMessage: string;
              switch (msgs.length) {
              case 1:
                errorName = msgs[0];
                break;
              case 2:
                errorName = msgs[0];
                errorMessage = msgs[1];
                if ('null' === errorMessage) {
                  errorMessage = null;
                }
                break;
              default:
                break;
              }
              const returnError: Error = {
                name: errorName,
                message: errorMessage,
              };
              promise.reject(returnError);
            } else {
              console.log(`System error[${error.Message}]`);
            }
          } else {
            console.log(`System error[${error.Code}|${error.Message}|${error.Data}]`);
          }
        } else {
          console.log('??????');
        }
      } else {
        // perhaps sever side send message
      }


  }
  private onNotificationHandler(notification: Notification): void {
    //
  }

  private fireOnDisconnect(): void {
    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private fireOnMessage(e: MessageEvent): void {


    for (let i = 0; i < this.onDisconnectListeners.length; i++) {
        this.onDisconnectListeners[i]();
    }
  }

  private getRequestID(): number {
    return ++this.requestID;
  }

}
